import logging
import sys
from logging.handlers import TimedRotatingFileHandler
from typing import Callable

import vatis.live_asr.config_variables as config_variables

FORMATTER = logging.Formatter("%(asctime)s — %(threadName)s — %(process)d — %(name)s — %(levelname)s — %(message)s")


def get_console_handler():
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)
    return console_handler


def get_file_handler():
    file_handler = TimedRotatingFileHandler(config_variables.LOGS_FILE, when='midnight')
    file_handler.setFormatter(FORMATTER)
    return file_handler


def create_logger(logger_name: str) -> logging.Logger:
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)  # better to have too much log than not enough
    logger.addHandler(get_console_handler())
    if config_variables.LOGS_FILE_ENABLED:
        logger.addHandler(get_file_handler())
    # with this pattern, it's rarely necessary to propagate the error up to parent
    logger.propagate = False
    return logger


"""
To configure the logging of this library, override this value 
"""
LOGGER_FACTORY_METHOD: Callable[[str], logging.Logger] = create_logger


def get_logger(logger_name: str) -> logging.Logger:
    return LOGGER_FACTORY_METHOD(logger_name)


def print_if_debugging(logger, msg, *args):
    if config_variables.DEBUG:
        logger.info(msg, args)



from vatis.asr_commons.config.audio import *
from vatis.live_asr.config_variables import *


class PerformanceConfig:
    """
    Performance configuration for a transcription stream. It determines how responsive the stream
    should be and how accurate the responses are
    """

    def __init__(self, frame_len: float, frame_overlap: float, buffer_offset: float):
        """
        Initialization method.
        The sum of :frame_len, :frame_overlap and :buffer_offset must be an integer number.
        General: a frame represent the samples extracted from the input signal during a period of 1 second.
        (e.g. for a signal sampled ad 16kHz, a frame is represented by 16000 consecutive samples from the signal)

        :param frame_len: Number of frames to be sent in each package.
        :param frame_overlap: Number of previously sent frames to be re-transcribed and refreshed in the transcript
        :param buffer_offset: Number of previously sent frames to be kept as a re-transcribe context.
        Those frames are not refreshed in the transcript
        """

        assert float(buffer_offset + frame_overlap + frame_len).is_integer()

        self._frame_len = frame_len
        self._frame_overlap = frame_overlap
        self._buffer_offset = buffer_offset

    @property
    def frame_len(self):
        return self._frame_len

    @property
    def frame_overlap(self):
        return self._frame_overlap

    @property
    def buffer_offset(self):
        return self._buffer_offset


SPEED_CONFIGURATION = PerformanceConfig(0.5, 1, 0.5)
ACCURACY_CONFIGURATION = PerformanceConfig(1, 2, 1)


class ConnectionConfig:
    AUTO_TIMEOUT: float = -1

    def __init__(self, reconnection_attempts: int = RECONNECTION_ATTEMPTS,
                 request_timeout: float = REQUEST_TIMEOUT, reconnection_delay: float = RECONNECTION_DELAY,
                 connection_timeout: float = CONNECTION_TIMEOUT):
        """
        :param reconnection_attempts: maximum number of initial connection or reconnection attempts
        :param request_timeout: maximum waiting time for a response from the server in seconds
        :param reconnection_delay: waiting time in seconds before performing a reconnection attempt
        :param connection_timeout: timeout for a reconnection attempt in seconds
        """
        self._reconnection_attempts = reconnection_attempts
        self._request_timeout = request_timeout
        self._reconnection_delay = reconnection_delay
        self._connection_timeout = connection_timeout

    @property
    def reconnection_attempts(self):
        return self._reconnection_delay

    @property
    def request_timeout(self):
        return self._request_timeout

    @property
    def reconnection_delay(self):
        return self._reconnection_delay

    @property
    def connection_timeout(self):
        return self._connection_timeout


DEFAULT_CONNECTION_CONFIG = ConnectionConfig()


class StreamConfig:
    """
    Configuration of the transcription stream
    """
    def __init__(self, language: Language, performance_config: PerformanceConfig = ACCURACY_CONFIGURATION,
                 connection_config: ConnectionConfig = DEFAULT_CONNECTION_CONFIG,
                 sample_rate: SampleRate = SampleRate.RATE_16000, channels: Channel = Channel.ONE,
                 bit_depth: BitDepth = BitDepth.BIT_16):
        """
        Initialization method

        :param language: language of the input signal
        :param performance_config: performance config for the stream
        :param connection_config: configuration for the connection between client and server
        :param sample_rate: signal sample rate
        :param channels: input signal sources (not implemented yet)
        :param bit_depth: augia signal bit depth
        """
        self._language = language
        self._performance_config = performance_config
        self._connection_config = connection_config
        self._sample_rate = sample_rate
        self._channels = channels
        self._bit_depth = bit_depth

    @property
    def performance_config(self) -> PerformanceConfig:
        return self._performance_config

    @property
    def connection_config(self) -> ConnectionConfig:
        return self._connection_config

    @property
    def language(self) -> Language:
        return self._language

    @property
    def sample_rate(self) -> SampleRate:
        return self._sample_rate

    @property
    def channels(self) -> Channel:
        return self._channels

    @property
    def bit_depth(self) -> BitDepth:
        return self._bit_depth



import os
from typing import Type


def get_env_var(key: str, default=None, throw_if_missing=False, dtype: Type = str):
    if key not in os.environ:
        if throw_if_missing:
            raise ValueError('Missing environment variable: ', key)
        else:
            return default

    return dtype(os.getenv(key))


# General ##############################################################################################################
DEBUG: bool = get_env_var('VATIS_ASR_CLIENT_DEBUG', default='False') == 'True'

# SocketIO Client ######################################################################################################
RECONNECTION_ATTEMPTS: int = get_env_var('VATIS_ASR_CLIENT_RECONNECTION_ATTEMPTS', default=6, dtype=int)
REQUEST_TIMEOUT: float = get_env_var('VATIS_ASR_CLIENT_REQUEST_TIMEOUT_SECONDS', default=15, dtype=float)
RECONNECTION_DELAY: float = get_env_var('VATIS_ASR_CLIENT_RECONNECTION_DELAY_SECONDS', default=5)
CONNECTION_TIMEOUT: float = get_env_var('VATIS_ASR_CLIENT_CONNECTION_TIMEOUT_SECONDS', default=10, dtype=float)

# Connection config ####################################################################################################
AUTHENTICATION_PROVIDER_URL: str = get_env_var('VATIS_ASR_CLIENT_AUTHENTICATION_PROVIDER_URL', default='https://vatis.tech/api/v1/asr-client/auth')

# Security #############################################################################################################
API_KEY: str = get_env_var('VATIS_ASR_CLIENT_API_KEY', throw_if_missing=True)

# Log ##################################################################################################################
LOGS_FILE: str = get_env_var('VATIS_ASR_CLIENT_LOGS_FILE', default='logs/app.logs', dtype=str)
LOGS_FILE_ENABLED: bool = get_env_var('VATIS_ASR_CLIENT_ENABLE_LOGS_FILE', default='False') == 'True'



from typing import Generator, Tuple, Callable
import wave

from vatis.live_asr.config.stream import StreamConfig

from vatis.live_asr.logging import get_logger


def file_generator(path: str, stream_config: StreamConfig) -> Generator[bytes, None, None]:
    """
    Creates a byte array generator the buffers the specified file.

    :param path: input file path
    :param stream_config: configuration object used to create the stream
    :return: byte array generator
    """
    CHUNK_SIZE: int = int(stream_config.sample_rate.value * stream_config.performance_config.frame_len)

    data: bytes = b''
    first_iter: bool = True

    with wave.open(path, 'rb') as wav_reader:
        while len(data) != 0 or first_iter:
            first_iter = False

            data = wav_reader.readframes(CHUNK_SIZE)

            yield data


def audio_source_generator(input_device_index, stream_config: StreamConfig, stop_condition: Callable[[int, float], bool] = None) -> Generator[bytes, None, None]:
    """
    Method for creating a byte array generator that buffers the input signal from the specified source.

    :param input_device_index: selected input device
    :param stream_config: configuration object used to create the stream
    :param stop_condition: predicate that receives the frame_count and starting_time in seconds relative to the Epoch
           and determines if the generator should stop. None means infinite streaming
    :return: byte array generator
    """
    import pyaudio
    from queue import Queue
    import time

    logger = get_logger('Input audio stream')

    audio = pyaudio.PyAudio()

    audio_format = pyaudio.get_format_from_width(stream_config.bit_depth.value)
    chunk_size: int = int(stream_config.sample_rate.value * stream_config.performance_config.frame_len)
    frame_count: int = 0
    starting_time: float = time.time()

    buffer: Queue[Tuple[bytes, int, dict]] = Queue()

    def new_audio_sample(data, samples, time_info, status_flags):
        buffer.put((data, samples, time_info))

        return data, pyaudio.paContinue

    stream = audio.open(format=audio_format, channels=stream_config.channels.value, stream_callback=new_audio_sample,
                        rate=stream_config.sample_rate.value, frames_per_buffer=chunk_size, input=True,
                        input_device_index=input_device_index)

    stream.start_stream()

    logger.info('Input audio stream started')

    stop_stream = False

    try:
        while not stop_stream:
            audio_sample: Tuple[bytes, int, dict] = buffer.get()
            frame_count += 1

            yield audio_sample[0]

            stop_stream = False if stop_condition is None else stop_condition(frame_count, starting_time)
    finally:
        stream.stop_stream()
        stream.close()

        audio.terminate()


def microphone_generator(stream_config: StreamConfig, stop_condition: Callable[[int, dict], bool] = None) -> Generator[bytes, None, None]:
    """
    Method for creating a byte array generator that buffers the input signal from the default input source (e.g. microphone).

    :param stream_config: configuration object used to create the stream
    :param stop_condition: predicate that receives the frame and time_info and determines if the generator should stop.
           None means infinite streaming
    :return: byte array generator
    """
    return audio_source_generator(input_device_index=None, stream_config=stream_config, stop_condition=stop_condition)


def print_input_devices():
    """
    Lists all the input device available
    """
    import pyaudio
    p = pyaudio.PyAudio()
    info = p.get_host_api_info_by_index(0)
    num_devices = info.get('deviceCount')

    for i in range(0, num_devices):
        if p.get_device_info_by_host_api_device_index(0, i).get('maxInputChannels') > 0:
            print("Input Device id ", i, " - ", p.get_device_info_by_host_api_device_index(0, i).get('name'))



from vatis.live_asr.config_variables import API_KEY, AUTHENTICATION_PROVIDER_URL
import requests
from requests import Response


class ClientAuthorization:
    def __init__(self, auth_token: str, service_host: str):
        self._auth_token = auth_token
        self._service_host = service_host

    @property
    def auth_token(self) -> str:
        return self._auth_token

    @property
    def service_host(self) -> str:
        return self._service_host


def get_auth_token() -> ClientAuthorization:
    response: Response = requests.get(
        AUTHENTICATION_PROVIDER_URL,
        params={'service': 'LIVE_ASR'},
        headers={'Authorization': 'Bearer ' + API_KEY}
    )

    if response.status_code != 200:
        raise ConnectionAbortedError(f'Bad status code {response.status_code}')

    parsed = response.json()

    return ClientAuthorization(**parsed)




from typing import Optional

from vatis.live_asr.stream.observer import LiveStreamObserver, ResponseMetadata
from vatis.asr_commons.domain.transcriber import TimestampedTranscriptionPacket
from vatis.asr_commons.live.headers import FINAL_FRAME_HEADER, FRAME_START_TIME_HEADER, FRAME_END_TIME_HEADER

import sys
from io import IOBase


class LoggingLiveStreamObserver(LiveStreamObserver):
    """
    Conventional stream observer for logging events
    """
    from logging import Logger

    def __init__(self, logger: Logger = None):
        super().__init__()
        self._logger = logger

    def on_connect(self):
        message = 'Stream connected'

        if self._logger is not None:
            self._logger.info(message)
        else:
            print(message)

    def on_disconnect(self):
        message = 'Stream disconnected'

        if self._logger is not None:
            self._logger.info(message)
        else:
            print(message)

    def on_response(self, packet: TimestampedTranscriptionPacket, metadata: ResponseMetadata):
        final_frame = 'FINAL' if packet.get_header(FINAL_FRAME_HEADER, default=False) else 'PARTIAL'
        message = f'{metadata.processing_time:.2f}s {final_frame:7s} : {packet.transcript}'

        if self._logger is not None:
            self._logger.info(message)
        else:
            print(message)

    def on_close(self):
        message = 'Stream closed'

        if self._logger is not None:
            self._logger.info(message)
        else:
            print(message)

    def on_transcription_completed(self):
        message = 'Transcription completed'

        if self._logger is not None:
            self._logger.info(message)
        else:
            print(message)


class FormattedLiveStreamObserver(LiveStreamObserver):
    """
    Utility class for formatting and redirecting the responses to an output stream
    """
    def __init__(self, out: Optional[IOBase] = None):
        if out is None:
            out = sys.stdout

        self._out = out

    def on_response(self, packet: TimestampedTranscriptionPacket, metadata: ResponseMetadata):
        if packet.get_header(FINAL_FRAME_HEADER, default=False):
            self._out.write('FINAL  : {0:.2f} -> {1:.2f}: {2}\n'.format(
                packet.get_header(FRAME_START_TIME_HEADER),
                packet.get_header(FRAME_END_TIME_HEADER),
                packet.transcript
            ))

        else:
            self._out.write('PARTIAL: {0:.2f} -> {1:.2f}: {2}\r'.format(
                packet.get_header(FRAME_START_TIME_HEADER),
                packet.get_header(FRAME_END_TIME_HEADER),
                packet.transcript
            ))
            self._out.flush()



from vatis.asr_commons.domain.transcriber import TimestampedTranscriptionPacket


class ResponseMetadata:
    def __init__(self, processing_time: int):
        self._processing_time = processing_time

    @property
    def processing_time(self) -> int:
        return self._processing_time


class LiveStreamObserver:
    def on_connect(self):
        """
        Called every time the stream connects
        """
        pass

    def on_disconnect(self):
        """
        Called every time the stream disconnects
        """
        pass

    def on_response(self, packet: TimestampedTranscriptionPacket, metadata: ResponseMetadata):
        """
        Callback for transcription response
        :param packet: the transcription response
        :param metadata: additional information about the response
        """
        pass

    def on_close(self):
        """
        Called when the stream closes and it will be no longer used
        """
        pass

    def on_transcription_completed(self):
        """
        Callback after a transcription is completed. This method will be called even if the stream was closed
        """
        pass



import base64
import json
import time
from queue import Queue
from threading import Condition
from typing import List, Generator, Callable, Optional
from threading import Thread, RLock

import socketio
from vatis.asr_commons.domain.transcriber import TimestampedTranscriptionPacket
from vatis.asr_commons.live.headers import PACKET_NUMBER_HEADER
from vatis.asr_commons.live.sio_topics import ASR_NAMESPACE, ASR_REQUEST_EVENT, ASR_RESULT_EVENT
from vatis.live_asr.stream.exception import TranscriptionError

from vatis.live_asr.config.stream import StreamConfig, ConnectionConfig
from vatis.live_asr.logging import get_logger
from vatis.live_asr.stream.observer import LiveStreamObserver, ResponseMetadata

logger = get_logger('live stream')


class LiveStream:
    WAITING_TIME_MULTIPLIER = 3

    """
    Class representing a transcription LiveStream. Each instance represents
    a different connection to the server.
    """
    def __init__(self, client: socketio.Client, config: StreamConfig):
        assert client.connected

        self._client = client
        self._config = config
        self._closed = False
        self._observers: List[LiveStreamObserver] = []
        self._last_package_sent_id = 0
        self._last_packet_received_id = 0
        self._all_packets_received: Condition = Condition()
        self._sent_timestamps = {}
        self._connected_condition = Condition()
        self._transcribing_lock = RLock()
        self._transcribing = False

        @client.event(namespace=ASR_NAMESPACE)
        def connect():
            with self._connected_condition:
                self._connected_condition.notifyAll()

            if not self._closed:
                for observer in self._observers:
                    try:
                        observer.on_connect()
                    except Exception as e:
                        logger.error('on_connected callback error', e)

        @client.event(namespace=ASR_NAMESPACE)
        def disconnect():
            self._last_package_sent_id = 0
            self._last_packet_received_id = 0

            if not self._closed:
                for observer in self._observers:
                    try:
                        observer.on_disconnect()
                    except Exception as e:
                        logger.error('on_disconnected callback error', e)

        @client.on(event=ASR_RESULT_EVENT, namespace=ASR_NAMESPACE)
        def on_message(message):
            parsed = json.loads(message)
            packet = TimestampedTranscriptionPacket(**parsed)

            self._last_packet_received_id = packet.get_header(PACKET_NUMBER_HEADER)

            metadata: ResponseMetadata = ResponseMetadata(
                processing_time=time.time() - self._sent_timestamps[self._last_packet_received_id]
            )

            self._sent_timestamps.pop(self._last_packet_received_id)

            with self._all_packets_received:
                if self._last_packet_received_id >= self._last_package_sent_id:
                    self._all_packets_received.notifyAll()

            if not self._closed:
                for observer in self._observers:
                    try:
                        observer.on_response(packet, metadata)
                    except Exception as e:
                        logger.error('on_disconnected callback error', e)

    @property
    def client(self) -> socketio.Client:
        return self._client

    @property
    def closed(self) -> bool:
        return self._closed

    @property
    def config(self) -> StreamConfig:
        return self._config

    def sid(self) -> str:
        return self._client.get_sid(namespace=ASR_NAMESPACE)

    def is_connected(self):
        return self._client.connected

    def send(self, data: bytes, received_callback=None, timeout: float = ConnectionConfig.AUTO_TIMEOUT):
        """
        Send signal frame for transcription.

        :param data: samples from the audio signal
        :param received_callback: confirmation of packet received callback
        :param timeout: waiting time in seconds to receive the transcription response for the given data frame
               once it was sent
        """
        if self._transcribing:
            raise TranscriptionError('Already transcribing')

        self._check_not_closed()

        self._feed(data, received_callback, timeout)

    def _feed(self, data: bytes, received_callback: Callable[..., None], timeout: Optional[float]):
        if timeout == ConnectionConfig.AUTO_TIMEOUT:
            timeout = self._config.performance_config.frame_len * LiveStream.WAITING_TIME_MULTIPLIER

        body = LiveStream._prepare_data(data)

        self._send_internal(body, received_callback=received_callback)

        self._last_package_sent_id += 1
        self._sent_timestamps[self._last_package_sent_id] = time.time()

        self._wait_for_completion(timeout)

    def _send_internal(self, body, received_callback):
        raise NotImplementedError('This is an abstraction method')

    @staticmethod
    def _prepare_data(data) -> dict:
        data_encoded = base64.b64encode(data)
        data_str = str(data_encoded, 'utf-8')

        return {'data': data_str}

    def _check_not_closed(self):
        if self._closed:
            raise ConnectionError('Stream disconnected')

    def add_observer(self, observer: LiveStreamObserver):
        """
        Registers a listener for stream events.

        :param observer: stream observer
        """
        self._check_not_closed()
        assert observer is not None

        self._observers.append(observer)

    def remove_observer(self, observer: LiveStreamObserver) -> bool:
        """
        Removes a stream observer

        :param observer: stream observer
        :return True if the element was deleted, False otherwise
        """
        assert observer is not None

        try:
            self._observers.remove(observer)
            return True
        except ValueError:
            return False

    def transcribe(self, data_generator: Generator[bytes, None, None], received_callback=None,
                   timeout: float = ConnectionConfig.AUTO_TIMEOUT) -> Thread:
        """
        Convenient method for transcribing a stream of data. The method is non-blocking.
        While a transcription is in process, no other transcription on this stream instance is allowed.

        :param data_generator: generator of data to be transcribed. The generated data must respect
                               the constrains specified in :send
        :param received_callback: callback for packet received by the server
        :param timeout: waiting time for processing a packet in seconds

        :return created thread
        """
        if self._transcribing:
            raise TranscriptionError('Already transcribing')

        def _transcribe():
            if self._transcribing_lock.acquire(False):
                self._transcribing = True

                try:
                    logger.info('Started transcription')

                    for data in data_generator:
                        self._feed(data, received_callback=received_callback, timeout=timeout)
                finally:
                    self._transcribing = False
                    self._transcribing_lock.release()
                    for observer in self._observers:
                        try:
                            observer.on_transcription_completed()
                        except Exception as e:
                            logger.error('on_disconnected callback error', e)
            else:
                logger.error('Already transcribing. Aborting transcription request')

        thread: Thread = Thread(target=_transcribe)
        thread.start()

        return thread

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self._closed = True

        if self._client.connected:
            self._client.disconnect()

        for observer in self._observers:
            try:
                observer.on_close()
            except Exception as e:
                logger.info('on_disconnected callback error', e)

    def _wait_for_completion(self, timeout: float = None, throw: bool = False) -> bool:
        """
        Blocks the current thread until all packets sent until this method call are transcribed.

        :param timeout: timeout to wait for completion
        :return False if method timed out, True otherwise
        """
        if self._last_packet_received_id < self._last_package_sent_id:
            with self._all_packets_received:
                packet_received = self._all_packets_received.wait(timeout)
                if not packet_received and throw:
                    raise TimeoutError(f'Timed out while waiting for the packet after {timeout} seconds')

                return packet_received

    def wait_for_transcription(self, timeout: float = None):
        """
        Blocks the current thread until the current transcription in process finishes. If no transcription is in process
        the method returns immediately
        :param timeout: waiting time for transcription to finish in seconds
        """
        if timeout is None:
            timeout = -1

        if self._transcribing_lock.acquire(True, timeout):
            self._transcribing_lock.release()
        else:
            raise TimeoutError(f'Timed out while waiting for the transcription after {timeout} seconds')

    def create_generator(self) -> Generator[TimestampedTranscriptionPacket, None, None]:
        """
        Creates a generator which yields all the response packets from the stream
        :return: responses generator
        """
        responses_queue = Queue()

        class ResponseObserver(LiveStreamObserver):
            def on_response(self, packet: TimestampedTranscriptionPacket, metadata: ResponseMetadata):
                responses_queue.put(packet)

            def on_transcription_completed(self):
                responses_queue.put(None)

        observer = ResponseObserver()

        self.add_observer(observer)

        while not self._closed:
            response = responses_queue.get()

            if response is not None:
                yield response
            else:
                break

        self.remove_observer(observer)
        del responses_queue


class SimpleLiveStream(LiveStream):
    """
    Simple implementation of a live stream. It tries to send the transcription packet directly
    to the server without connection checking or retrying mechanism.
    """
    def __init__(self, client: socketio.Client, config: StreamConfig):
        super().__init__(client, config)

    def _send_internal(self, body, received_callback):
        self._client.emit(
            event=ASR_REQUEST_EVENT,
            data=body,
            namespace=ASR_NAMESPACE,
            callback=received_callback
        )


class BlockingLiveStream(LiveStream):
    """
    Live Stream implementation that blocks the :send method if the client is not currently connected until
    it's connected again or the timeout is reached.
    """
    def __init__(self, client: socketio.Client, config: StreamConfig):
        super().__init__(client, config)

    def _send_internal(self, body, received_callback):
        if not self._client.connected and not self._closed:
            connection_timeout: float = (self._config.connection_config.connection_timeout + self._config.connection_config.reconnection_delay) * \
                                         self._config.connection_config.reconnection_attempts
            with self._connected_condition:
                if not self._client.connected:
                    logger.warning(f'Client disconnected, waiting for reconnection up to {connection_timeout} seconds')
                    if not self._connected_condition.wait(connection_timeout):
                        raise TimeoutError(f'Timed out while waiting for the packet after {connection_timeout} seconds')

        self._client.emit(
            event=ASR_REQUEST_EVENT,
            data=body,
            namespace=ASR_NAMESPACE,
            callback=received_callback
        )



from vatis.live_asr.config.stream import PerformanceConfig
from vatis.live_asr.config.stream import StreamConfig
from vatis.live_asr.config.stream import SPEED_CONFIGURATION
from vatis.live_asr.config.stream import ACCURACY_CONFIGURATION

__all__ = (
    "PerformanceConfig",
    "StreamConfig",
    "SPEED_CONFIGURATION",
    "ACCURACY_CONFIGURATION"
)



import io

from setuptools import find_namespace_packages, setup

import os

__name__ = 'vatis_asr_client'
__tag__ = '1.1.1'
__short_description__ = 'Python implementation of the client for Vatis\'s ASR services'
__download_url__ = 'https://gitlab.com/vatistech/asr-client-python/-/archive/{__tag__}/asr-client-python-{__tag__}.zip'\
    .format(__tag__=__tag__)

# Should be one of:
# 'Development Status :: 3 - Alpha'
# 'Development Status :: 4 - Beta'
# 'Development Status :: 5 - Production/Stable'
__release_status__ = "Development Status :: 4 - Beta"


def req_file(filename, folder="requirements"):
    with open(os.path.join(folder, filename)) as f:
        content = f.readlines()
    # you may also want to remove whitespace characters
    # Example: `\n` at the end of each line
    return [x.strip() for x in content]


install_requires = req_file('requirements.txt')
extras_require = {
    'remote': req_file('requirements_remote.txt'),
    'standalone': req_file('requirements_standalone.txt')
}

package_root = os.path.abspath(os.path.dirname(__file__))

readme_filename = os.path.join(package_root, "README.md")
with io.open(readme_filename, encoding="utf-8") as readme_file:
    readme = readme_file.read()

packages = find_namespace_packages(include=['vatis.*'])

namespaces = ['vatis']

setup(
    name=__name__,
    version=__tag__,
    description=__short_description__,
    long_description_content_type='text/markdown',
    long_description=readme,
    url='https://gitlab.com/vatistech/asr-client-python',
    download_url=__download_url__,
    maintainer='VATIS TECH',
    maintainer_email='founders@vatis.tech',
    packages=packages,
    namespace_packages=namespaces,
    include_package_data=True,
    install_requires=install_requires,
    extras_require=extras_require,
    zip_safe=False,
    python_requires=">=3.6",
    platforms="Posix; MacOS X; Windows",
    license='Apache Software License',
    classifiers=[
        __release_status__,
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Internet"
    ]
)



import socketio

from vatis.live_asr.config_variables import *
from vatis.live_asr.service import auth as auth_service
from vatis.live_asr.config.stream import StreamConfig

from vatis.asr_commons.live.headers import *
from vatis.asr_commons.live.sio_topics import *

from vatis.live_asr.logging import get_logger
from vatis.live_asr.service.auth import ClientAuthorization
from vatis.live_asr.stream.transcription import LiveStream, SimpleLiveStream, BlockingLiveStream

logger = get_logger('SIO Connection factory')


def create_stream(stream_config: StreamConfig, stream_type: str = 'BLOCKING') -> LiveStream:
    """
    Factory method for creating a LiveStream.

    :param stream_config: transcription stream configuration
    :param stream_type: stream to be created. Possible values are: SIMPLE, BLOCKING
    :return: LiveStream instance
    """
    sio = _create_sio_client(stream_config)

    if stream_type == 'SIMPLE':
        return SimpleLiveStream(sio, stream_config)
    elif stream_type == 'BLOCKING':
        return BlockingLiveStream(sio, stream_config)
    else:
        raise ValueError(f'Unsupported value {stream_type}')


def _create_sio_client(stream_config: StreamConfig) -> socketio.Client:
    sio: socketio.Client = socketio.Client(logger=DEBUG, engineio_logger=DEBUG, ssl_verify=False,
                                           request_timeout=stream_config.connection_config.request_timeout,
                                           reconnection_attempts=stream_config.connection_config.reconnection_attempts,
                                           reconnection_delay=stream_config.connection_config.reconnection_delay,
                                           reconnection=True)

    authorization: ClientAuthorization = auth_service.get_auth_token()
    auth_token: str = authorization.auth_token

    if not auth_token.startswith('Bearer'):
        auth_token = 'Bearer ' + auth_token

    service_url = f'/live/{stream_config.language.value}/transcribe'
    url = authorization.service_host + service_url

    headers = {
        'Authorization': auth_token,
        FRAME_LEN_HEADER: str(stream_config.performance_config.frame_len),
        FRAME_OVERLAP_HEADER: str(stream_config.performance_config.frame_overlap),
        BUFFER_OFFSET_HEADER: str(stream_config.performance_config.buffer_offset)
    }

    logger.debug(f'Connecting to {url} with headers {headers}')

    connection_attempts = 0

    while not sio.connected and connection_attempts < stream_config.connection_config.reconnection_attempts:
        try:
            logger.info('Connecting to ASR service. Attempt %d', connection_attempts)
            sio.connect(url=url + '/socket.io',
                        namespaces=[ASR_NAMESPACE],
                        socketio_path=service_url + '/socket.io',
                        transports=['websocket'],
                        wait=True,
                        wait_timeout=stream_config.connection_config.connection_timeout,
                        headers=headers)

            logger.info("Connection to ASR service established")
        except Exception as e:
            logger.error("Connection error", e)
            connection_attempts += 1

    if connection_attempts >= stream_config.connection_config.reconnection_attempts:
        raise ConnectionError(f'Connection to {url} failed after {connection_attempts} attempts')

    return sio




from vatis.live_asr.stream.factory import create_stream
from vatis.live_asr.stream.observer import LiveStreamObserver, ResponseMetadata

__all__ = (
    "LiveStreamObserver",
    "ResponseMetadata",
    "create_stream"
)

